/*
 * net.c
 *
 * A library of network routines
 *
 * Kekoa Proudfoot 12/31/96
 */

/*****************************************************************************
 * Include files
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <sys/time.h>
#endif
#include <ctype.h>
#include <errno.h>
#include <string.h>

#include "net.h"
#include "endian.h"
#include "debug.h"
#include "hexdump.h"

#ifdef WIN32
void pwserror(char *);
#endif

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

/*****************************************************************************
 * Configuration macros
 *****************************************************************************/

#define NET_NUM_BIND_ATTEMPTS  50

#define NET_FIRST_RANDOM_PORT  1024
#define NET_NUM_RANDOM_PORTS   512

/*****************************************************************************
 * Global variables
 *****************************************************************************/

static fd_set global_fds;

/*****************************************************************************
 * Interface function definitions
 *****************************************************************************/

void
net_init(void)
{
#ifdef WIN32
  WSADATA wsaData;

  if( WSAStartup(0x202, &wsaData) == SOCKET_ERROR ) {
    pwserror("WSAStartup");
    WSACleanup();
	exit(1);
  }
#endif
}

void
net_listen(int s)
{
  if( listen(s, 5) < 0 ) {
    perror("listen");
    exit(1);
  }
}

void
net_close_fd(int fd)
{
    FD_CLR((unsigned int)fd, &global_fds);
}

void
net_open_fd(int fd)
{
    FD_SET(fd, &global_fds);
}

int
net_open_socket(int type)
{
    int s;

    if ((s = socket(AF_INET, type, 0)) < 0) {
#ifdef WIN32
        pwserror("socket");
#else
        perror("socket");
#endif
	exit(1);
    }

    FD_SET(s, &global_fds);

    return s;
}

void
net_close_socket(int s)
{
#ifdef WIN32
    if (s && closesocket(s) < 0) {
      pwserror("closesocket");
      exit(1);
    }
#else
    if (s && close(s) < 0) {
      perror("close");
      exit(1);
    }
#endif

    FD_CLR((unsigned int)s, &global_fds);
}

void
net_connect_socket(int s, struct sockaddr_in *a, short p)
{
    a->sin_port = (unsigned short)htons(p);
    if (connect(s, (struct sockaddr *)a, sizeof(struct sockaddr_in)) < 0) {
#ifdef WIN32
        pwserror("connect");
#else
        perror("connect");
#endif
	exit(1);
    }
}

void
net_bind_socket(int s, struct sockaddr_in *a, short p)
{
    a->sin_port = (unsigned short)htons(p);
    if (bind(s, (struct sockaddr *)a, sizeof(struct sockaddr_in)) < 0) {
#ifdef WIN32
        pwserror("bind");
#else
        perror("bind");
#endif
	exit(1);
    }
}

void
net_bind_socket_random(int s, struct sockaddr_in *a, short *p)
{
    int i;

    for (i = 0; i < NET_NUM_BIND_ATTEMPTS; i++) {
	*p = NET_FIRST_RANDOM_PORT + rand() % NET_NUM_RANDOM_PORTS;
	a->sin_port = htons(*p);
	if (bind(s, (struct sockaddr *)a, sizeof(struct sockaddr_in)) >= 0)
	    return;
    }

#ifdef WIN32
    pwserror("bind");
#else
    perror("bind");
#endif
    exit(1);
}

char *
net_address_to_name(struct sockaddr_in *a)
{
    char           *bufp;
    static char     buf[1024];
    struct hostent *host;

    /* Possible optimization would be to cache addresses and names */

    if ((host = gethostbyaddr((char *)&a->sin_addr, sizeof(struct in_addr),
			      AF_INET)) == NULL) {
	sprintf(buf, "%s:%d", inet_ntoa(a->sin_addr), ntohs(a->sin_port));
    }
    else {
	sprintf(buf, "%s:%d", host->h_name, ntohs(a->sin_port));

	for (bufp = buf; *bufp; bufp++) {
	    *bufp = tolower(*bufp);
	}
    }

    return buf;
}

struct sockaddr_in
net_name_to_address(char *name)
{
    char               *buf;
    char               *port;
    struct hostent     *host;
    struct sockaddr_in  addr;

    memset(&addr, 0, sizeof(struct sockaddr_in));

    buf = strdup(name);
    strtok(buf, ":");
    if ((port = strtok(NULL, "")) != NULL) {
        addr.sin_port = atoi(port);
        if (addr.sin_port <= 0 || addr.sin_port >= (1<<16))
            debug_fatal("net_name_to_address: %s: invalid port number", port);
    }
    else {
        addr.sin_port = 0;
    }

    /*First see if 'buf' is in the form 1.2.3.4, if it is,
      just use that as the ipaddr, if not, then do a gethostbyname*/
    addr.sin_addr.s_addr = inet_addr(buf);

    if( addr.sin_addr.s_addr == INADDR_NONE) {
       if ((host = gethostbyname(buf)) != NULL) {
          addr.sin_addr   = *(struct in_addr *)host->h_addr;
       } else {
          debug_fatal("%s: couldn't determine ip address", buf);
       }
    }

    addr.sin_family = AF_INET;
    addr.sin_port   = htons(addr.sin_port);

    free(buf);

    return addr;
}

char *
net_name_to_numbers(char *name)
{
    static char         buf[1024];
    struct sockaddr_in  addr;

    addr = net_name_to_address(name);
    
    sprintf(buf, "%s:%d", inet_ntoa(addr.sin_addr), ntohs(addr.sin_port));

    return buf;
}

int
net_send(int s, packet_t *p)
{
    int len;

    packet_host_to_network(p);

    len = read_big_short(&p->header.length);

    if (send(s, (u_char *)p, len, 0) != len) {
        return 0;
    }

    packet_network_to_host(p);

    return 1;
}

int
net_recv(int s, packet_t *p)
{
    int len;

    if ((len = recv(s, (u_char *)p, sizeof(packet_t), 0)) < 0) {
        return 0;
    }

    if (len != read_big_short(&p->header.length)) {
	p->header.type   = PACKET_UNDEFINED;
	p->header.length = 0;
    }

    packet_network_to_host(p);

    return 1;
}

void
net_send_to(int s, struct sockaddr_in *ain, packet_t *p)
{
    int len;
    struct sockaddr *a = (struct sockaddr *)ain;

    packet_host_to_network(p);

    len = read_big_short(&p->header.length);

    if (sendto(s, (u_char *)p, len, 0, a, sizeof(struct sockaddr_in)) != len) {
#ifdef WIN32
      pwserror("sendto");
#else
      perror("sendto");
#endif
      exit(1);
    }

    packet_network_to_host(p);
}

void
net_recv_from(int s, struct sockaddr_in *ain, packet_t *p)
{
    int len, addrlen;
    struct sockaddr *a = (struct sockaddr *)ain;

    addrlen = sizeof(struct sockaddr_in);

    if ((len = recvfrom(s, (u_char *)p, sizeof(packet_t), 0, a, &addrlen)) < 0) {
#ifdef WIN32
        pwserror("recvfrom");
        exit(1);
#else
	if (errno != ECONNREFUSED) {
	    perror("recvfrom");
	    exit(1);
	}
#endif
    }

    if (len != read_big_short(&p->header.length)) {
	p->header.type   = PACKET_UNDEFINED;
	p->header.length = 0;
    }

    packet_network_to_host(p);
}

int
net_check_for_input(int s)
{
    return net_wait_for_input(s, 0.0);
}

int
net_wait_for_input(int s, float t)
{
    fd_set          fds;
    struct timeval  timeout;

    /*Hmm... QSmack uses STDIN which has a fd of zero*/
/*    if (!s) return 0;*/

    timeout.tv_sec  = (int)(t);
    timeout.tv_usec = (int)((t - (int)(t)) * 1e6 + 0.5);

    FD_ZERO(&fds);
    FD_SET(s, &fds);

    if (select(FD_SETSIZE, &fds, NULL, NULL, &timeout) < 0) {
#ifdef WIN32
      pwserror("select");
#else
      perror("select");
#endif
	exit(1);
    }

    return FD_ISSET(s, &fds);
}

int
net_wait_for_any_input(float t)
{
    int             count;
    fd_set          fds;
    struct timeval  timeout;

    timeout.tv_sec  = (int)(t);
    timeout.tv_usec = (int)((t - (int)(t)) * 1e6 + 0.5);

    memcpy(&fds, &global_fds, sizeof(global_fds));
    
    if ((count = select(FD_SETSIZE, &fds, NULL, NULL, &timeout)) < 0) {
#ifdef WIN32
      pwserror("select");
#else
      perror("select");
#endif
      exit(1);
    }

    return (count > 0);
}

#ifdef WIN32
void
pwserror(char *str)
{
  int i;

  fprintf(stderr,"%s: ", str);

  i = WSAGetLastError();
  switch ( i ) {
  case WSAENETDOWN:
    fprintf(stderr, "the network is down.\n");
    break;
  case WSAEINTR:
    fprintf(stderr, "a blocking call was cancelled.\n");
    break;
  case WSAEINPROGRESS:
    fprintf(stderr, "a blocking call in progress.\n");
    break;
  case WSAEADDRINUSE:
    fprintf(stderr, "address already in use.\n");
    break;
  case WSAEADDRNOTAVAIL:
    fprintf(stderr, "address not available.\n");
    break;
  case WSAECONNREFUSED:
    fprintf(stderr, "connection refused.\n");
  case WSAENOBUFS:
    fprintf(stderr, "no network buffer space.\n");
  case WSAENETRESET:
    fprintf(stderr, "connection unexpectedly reset.\n");
    break;
  default:
    fprintf(stderr, "unknown network error %d.\n", i);
    break;
  }
}
#endif
